/*
  IOT Task - Learning Assistant

  This sketch manages three modes: STANDBY, STUDY, and BREAK, using an ESP32 microcontroller.
  The system controls an LED strip, a single LED, an IR remote, and interacts with the Blynk app and Spotify via webhooks.

  The modes:
  1. STANDBY:
     - LED is off.
     - Pressing the OK button starts the study-break cycle (currently set to 45 minutes study and 15 minutes break).

  2. STUDY:
     - Turns off the TV at the start of the study session.
     - LED is on.
     - Plays study music on Spotify via a webhook (MAKE).
     - LED strip shows the remaining time (blue color).
     - Pressing the OK button switches to BREAK mode.
     - Checks every 5 minutes if the temperature is within the set range and controls the AC accordingly.
     - Sends a notification that study mode is starting via a webhook (MAKE).


  3. BREAK:
     - LED is off.
     - LED strip shows the remaining time (red color).
     - Sends a notification that break mode is starting via a webhook (MAKE).
     - Sends a notification about the remaining break time via a webhook (MAKE).
     - Stops the relaxing music.

  The circuit:
  * Input components:
    - IR receiver connected to pin 34.
    - DHT22 temperature sensor connected to pin 16.

  * Output components:
    - Single LED connected to pin 2.
    - LED strip connected to pin 26.
    - IR sender connected to pin 12.
*/

//BLYNK credentials
#define BLYNK_TEMPLATE_ID           "Your_Template_ID"
#define BLYNK_TEMPLATE_NAME         "Your_Template_Name"
#define BLYNK_AUTH_TOKEN            "Your_Auth_Token"
#define BLYNK_PRINT Serial

// Required Libraries
#include <Arduino.h>
#include <WiFi.h>
#include <WiFiClient.h>
#include <WiFiClientSecure.h>
#include <Adafruit_NeoPixel.h>
#include <IRremote.h>
#include <DHT.h>
#include <HTTPClient.h>
#include <BlynkSimpleEsp32.h>

// Output pins
#define LED_PIN 2
#define LED_STRIP_PIN 26
#define NUM_LEDS 12
#define IR_RECEIVE_PIN 34
#define IR_SEND_PIN 12

// WiFi Credentials
const char* ssid = "Your_SSID";
const char* pass = "Your_Password";

// LED Strip Configuration
const int numLeds = 12;
Adafruit_NeoPixel strip = Adafruit_NeoPixel(numLeds, LED_STRIP_PIN, NEO_GRB + NEO_KHZ800);

// Create IR receiver object
IRrecv irrecv(IR_RECEIVE_PIN);
decode_results results;

// Create IR sender object
IRsend irsend;

//initialize the DHT
#define DHTPIN 16
#define DHTTYPE DHT22
DHT dht(DHTPIN, DHTTYPE);

// Blynk Virtual Pins
#define VPIN_MODE V1
#define VPIN_TIMER V2
#define VPIN_BREAK_COUNT V3
#define VPIN_LEARNING_TIME V4

// Define IR remote button codes
const unsigned long BUTTON_0 = 0xAD52FF00; // code for '0' button
const unsigned long BUTTON_1 = 0xE916FF00; // code for '1' button
const unsigned long BUTTON_2 = 0xE619FF00; // code for '2' button
const unsigned long BUTTON_3 = 0xF20DFF00; // code for '3' button
const unsigned long BUTTON_4 = 0xF30CFF00; // code for '4' button
const unsigned long BUTTON_5 = 0xE718FF00; // code for '5' button
const unsigned long BUTTON_6 = 0xA15EFF00; // code for '6' button
const unsigned long BUTTON_7 = 0xF708FF00; // code for '7' button
const unsigned long BUTTON_8 = 0xE31CFF00; // code for '8' button
const unsigned long BUTTON_9 = 0xA55AFF00; // code for '9' button
const unsigned long BUTTON_UP = 0xB946FF00; // code for 'up' button
const unsigned long BUTTON_DOWN = 0xEA15FF00; // code for 'down' button
const unsigned long BUTTON_LEFT = 0xBB44FF00; // code for 'LEFT' button
const unsigned long BUTTON_RIGHT = 0xBC43FF00; // code for 'RIGHT' button
const unsigned long BUTTON_HASHTAG = 0xB54AFF00; // code for 'HASHTAG' button
const unsigned long BUTTON_STAR = 0xBD42FF00; // code for 'STAR' button
const unsigned long BUTTON_OK = 0xBF40FF00; // code for 'OK' button

const unsigned long BUTTON_TV = 0xF708FB04; // code for 'TV turn off' button
const unsigned long BUTTON_AC_OFF = 0x86C0051; // code for 'AC turn off' button
const unsigned long BUTTON_AC_ON = 0x8600145; // code for 'AC turn on' button

unsigned long wifipreviousMillis = 0; //for continuous Multitasking programming
unsigned long mode_start_time_millis = 0; //for continuous Multitasking programming

BlynkTimer timer;

int mode = 0; //0 - stand by    1 - study     2 - break
bool modeChanged = false;

int study_minutes = 45; //how much time for a study session 
int study_count = 0; //study sessions count
int total_minutes_studied = 0; //total study time in minutes
int total_break_time = 0; //total break time in minutes

int break_minutes = 15; //how much time for a break
int break_count = 0; //breaks count

int cold_temp = 21; //minumum temperature in the room 
int hot_temp = 25;  //maximum temperature in the room

// Webhook URLs for Make (Integromat)
const char* studyModeUrl = "Your_Study_Mode_Webhook_URL";
const char* breakModeUrl = "Your_Break_Mode_Webhook_URL";

// Function to send HTTP requests
void sendHttpRequest(const char* url) {
  if (WiFi.status() == WL_CONNECTED) {
    HTTPClient http;
    http.begin(url);
    int httpResponseCode = http.GET();
    Serial.print("HTTP Response code: ");
    Serial.println(httpResponseCode);
    http.end();
  } else {
    Serial.println("WiFi Disconnected");
  }
}

void rainbowlight()
{
  for(long firstPixelHue = 0; firstPixelHue < 5*65536; firstPixelHue += 256) 
  {
    for(int i=0; i<12; i++) // Only iterate over 12 LEDs
    { 
      // Hue of the pixel, changing over time.
      int pixelHue = firstPixelHue + (i * 65536L / strip.numPixels());
      // Set pixel color using wheel function
      strip.setPixelColor(i, strip.gamma32(strip.ColorHSV(pixelHue)));
    }
    strip.show();
  }
}

//turns off AC
void AC_Off() {
  irsend.sendLG(0x86, 0x14, 0);
  Serial.println("Turned off AC");
}

//turns on AC
void AC_On() {
  irsend.sendLG(0x86, 0x14, 0);
  Serial.println("Turned on AC");
}

//turns off TV
void TV_Off() {
  irsend.sendNEC(0x00, 0x51, 0);
  Serial.println("Turned off TV");
}

void colorWipe(uint32_t color, int wait) {
  for (int i = 0; i < strip.numPixels(); i++) { // For each pixel in strip...
    strip.setPixelColor(i, color);         //  Set pixel's color (in RAM)
    strip.show();                          //  Update strip to match
    delay(wait);                           //  Pause for a moment
  }
}

void updateBlynk() {
  String currentMode = (mode == 0) ? "STANDBY" : (mode == 1) ? "STUDY" : "BREAK";
  Blynk.virtualWrite(VPIN_MODE, currentMode);

  unsigned long elapsedMillis = millis() - mode_start_time_millis;
  int totalMinutes = (mode == 1) ? study_minutes : break_minutes;
  int remainingMinutes = totalMinutes - (elapsedMillis / 60000);
  Blynk.virtualWrite(VPIN_TIMER, remainingMinutes);

  Blynk.virtualWrite(VPIN_BREAK_COUNT, total_break_time); // Total break time in minutes
  Blynk.virtualWrite(VPIN_LEARNING_TIME, total_minutes_studied); // Total study time in minutes
}

void studyMode()
{
  unsigned long currentMillis = millis(); //checking current time for Multitasking

  Serial.print("breaked for:");
  Serial.print((currentMillis - mode_start_time_millis)/60/1000);
  Serial.println("  minutes");
  Serial.println();

  total_minutes_studied += (currentMillis - mode_start_time_millis)/1000/60;
  mode_start_time_millis = currentMillis; //logging the start time
  colorWipe(strip.Color(  0,   0, 255), 50); // Blue LED strip
  digitalWrite(LED_PIN, HIGH); //ligths up the LED light
  mode = 1;
  study_count++; //increasing the number of studying sessions
  TV_Off(); //turns off TV
  //sendHttpRequest(playMusicUrl); // Play music
  sendHttpRequest(studyModeUrl); // Notify Make about study mode start


  Serial.println("changed to study mode");
  updateBlynk();
}

void breakMode()
{
  unsigned long currentMillis = millis(); //checking current time for Multitasking

  Serial.print("studied for:  ");
  Serial.print((currentMillis - mode_start_time_millis)/60/1000);
  Serial.println("  minutes");
  Serial.println();

  total_break_time += (currentMillis - mode_start_time_millis)/1000/60;
  mode_start_time_millis = currentMillis; //logging the start time
  colorWipe(strip.Color(255,   0,   0), 50); // Red LED strip
  digitalWrite(LED_PIN, LOW); //shuts down the LED light

  sendHttpRequest(breakModeUrl); // Notify Make about break mode start

  mode = 2;
  break_count++;
  Serial.println("changed to break mode");
  updateBlynk();
}

//initialize stand-by mode
void standbyMode()
{
  colorWipe(strip.Color(  0, 0, 0), 50);//turns off LED strip
  digitalWrite(LED_PIN, LOW); //shuts down the LED light
  mode = 0;
  Serial.println("changed to standby mode");
  updateBlynk();
}

void changeMode()
{
  Serial.println("change mode function");
  unsigned long currentMillis = millis(); //checking current time for Multitasking
  if(mode == 1 && (currentMillis - mode_start_time_millis)/60/1000 > study_minutes) //if been studying for 45 minutes
  {
    Serial.println("changing to break mode");
    breakMode();
  }
  else if(mode == 2 && (currentMillis - mode_start_time_millis) > break_minutes * 60 * 1000) //if been on break for 15 minutes
  {
    Serial.println("changing to study mode");
    studyMode();
  }
}

//check the temperature, 
//if too hot  - turns on AC
//if too cold - turns off AC
//hot_temp and cold_temp initialized as global variables
void check_temperature()
{
  float temperature = dht.readTemperature();
  Serial.print("Temperature: ");
  Serial.println(temperature);

// Send temperature to Blynk
  Serial.print("Sending temperature to Blynk: ");
  Serial.println(temperature);
  Blynk.virtualWrite(V0, temperature);

  if(temperature > hot_temp)
  {
    AC_On();
  }
  if(temperature < cold_temp)
  {
    AC_Off();
  }
}

// Updates the LED strip to indicate the remaining time for the current study or break session.
// Lights up the LEDs in blue for study mode and red for break mode proportionally to the time left.
void updateLedStrip() {
  unsigned long currentMillis = millis();
  unsigned long elapsedMillis = currentMillis - mode_start_time_millis;
  int totalMinutes = (mode == 1) ? study_minutes : break_minutes;
  int remainingMinutes = totalMinutes - (elapsedMillis / 60000);

  // Calculate the number of LEDs to light up
  int ledsToLight = (remainingMinutes * numLeds) / totalMinutes;

  // Set the color based on the mode
  uint32_t color = (mode == 1) ? strip.Color(0, 0, 255) : strip.Color(255, 0, 0);

  // Update the LED strip
  for (int i = 0; i < numLeds; i++) {
    if (i < ledsToLight) {
      strip.setPixelColor(i, color);
    } else {
      strip.setPixelColor(i, 0); // Turn off the LED
    }
  }
  strip.show();
}

//setup function
void setup() 
{
  Serial.begin(115200); //debugging output

  Blynk.begin(BLYNK_AUTH_TOKEN, ssid, pass);

  pinMode(LED_PIN, OUTPUT); //single led 

  strip.begin();// Initialize the LED strip
  strip.show(); // Initialize all pixels to 'off'
  strip.setBrightness(50);//originally was 150 

  IrReceiver.begin(IR_RECEIVE_PIN, ENABLE_LED_FEEDBACK); // Start the receiver
  irsend.begin(IR_RECEIVE_PIN); // Initialize the IR sender

  dht.begin();//start DHT reciever

  timer.setInterval(10000L, changeMode); // Call changeMode every 10 second to check if a mode needs to be changed
  timer.setInterval(300000L, check_temperature); // Call check_temperature every 5 minutes (300000 milliseconds) 
  timer.setInterval(60000L, updateLedStrip); // Call updateLedStrip every minute
  timer.setInterval(1000L, updateBlynk); // Call updateBlynk every second

}

//loop function
void loop() 
{
  unsigned long currentMillis = millis(); //checking current time for Multitasking
  if (currentMillis - wifipreviousMillis >= 60 * 1000)
  {
    wifipreviousMillis = currentMillis;
    if(WiFi.status() != WL_CONNECTED) //reconnects to the wifi if it was disconnected
    {
      Serial.println("Reconnecting to wifi");
      Blynk.begin(BLYNK_AUTH_TOKEN, ssid, pass);
    }
  }

  Blynk.run();
  timer.run(); //activates Interval functions - changeMode & check_temperature (by clock)

  if (IrReceiver.decode()) 
  {
    Serial.println(IrReceiver.decodedIRData.decodedRawData, HEX);

    if(IrReceiver.decodedIRData.decodedRawData == BUTTON_OK)
    {
      Serial.println("OK was pressed");
      if(mode == 0)
      {
        Serial.println("lets start studying");
        studyMode();
      }
      else if(mode == 1)
      {
        Serial.println("lets take a break");
        breakMode();
      }
      else if(mode == 2)
      {
        Serial.println("lets continue studying");
        studyMode();
      }
      
    }

    if(IrReceiver.decodedIRData.decodedRawData == BUTTON_0)
    {
      Serial.println("'zero' button was pressed");
      if(mode == 1 || mode == 2)
      {
        Serial.println("We're done studying");
        standbyMode();
      }
    }

    IrReceiver.resume(); // Enable receiving of the next value
  }
}